#! /usr/bin/env python
#
# Copyright INRA-URGI 2009-2013
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
# 
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
from optparse import OptionParser
from commons.core.parsing.BlastParser import BlastParser
from commons.core.parsing.FastaParser import FastaParser
from SMART.Java.Python.structure.Transcript import Transcript
from SMART.Java.Python.structure.Mapping import Mapping
from SMART.Java.Python.misc.RPlotter import RPlotter
from SMART.Java.Python.misc.UnlimitedProgress import UnlimitedProgress
from SMART.Java.Python.misc.Progress import Progress

BIN_SIZE  = 10
QUERY     = 0
REF       = 1
TYPES     = (QUERY, REF)
TYPE_NAME = {QUERY: "query", REF: "ref"}

class Evaluator(object):

	def __init__(self, verbosity):
		self.verbosity  = verbosity
		self.plots      = False
		self.sizes      = {}
		self.names      = {}
		self.categories = False

	def setInputFile(self, fileName):
		self.parser = BlastParser(fileName, self.verbosity)

	def setReferenceFile(self, fileName):
		self.setFile(fileName, REF)

	def setQueryFile(self, fileName):
		self.setFile(fileName, QUERY)

	def setFile(self, fileName, type):
		fastaParser      = FastaParser(fileName, self.verbosity)
		self.names[type] = [te.split("#")[0] for te in fastaParser.getRegions()]
		self.sizes[type] = dict([te.split("#")[0], fastaParser.getSizeOfRegion(te)] for te in fastaParser.getRegions())
		if type == REF and self.categories:
			self.categories = {}
			for teName in fastaParser.getRegions():
				key, value = teName.split("#")
				self.categories[key] = value

	def setOutputFile(self, fileName):
		self.outputFileName = fileName

	def selectTes(self, fileName):
		tes    = []
		handle = open(fileName)
		for line in handle:
			tes.append(line.strip())
		for te in self.names[REF]:
			if te not in tes:
				del self.names[REF][te]
				del self.sizes[REF][te]

	def setPlots(self, boolean):
		self.plots = boolean

	def setCategories(self, boolean):
		self.categories = boolean

	def _initializeData(self):
		self.coverage   = {}
		self.matches    = {}
		self.depthes    = {}
		self.nbParts    = {}
		self.fullLength = {}
		for type in TYPES:
			self.coverage[type]    = {}
			self.matches[type]     = {}
			self.nbParts[type]     = {}
			self.fullLength[type]  = {}
			for name in self.names[type]:
				self.coverage[type][name]   = {}
				self.matches[type][name]    = set()
				self.nbParts[type][name]    = 0
				self.fullLength[type][name] = {}

	def _addMapping(self, mapping):
		refInt    = mapping.targetInterval
		ref       = refInt.getChromosome().split("#")[0]
		refStart  = refInt.getStart()-1
		refEnd    = refInt.getEnd()-1
		for pos in range(refStart, refEnd+1):
			self.coverage[REF][ref][pos] = self.coverage[REF][ref].get(pos, 0) + 1
		queryInt   = mapping.queryInterval
		query      = queryInt.getName()
		queryStart = queryInt.getStart()-1
		queryEnd   = queryInt.getEnd()-1
		for pos in range(queryStart, queryEnd+1):
			self.coverage[QUERY][query][pos] = self.coverage[QUERY][query].get(pos, 0) + 1
		self.fullLength[REF][ref][query]   = self.fullLength[REF][ref].get(query, 0)   + (refEnd - refStart + 1)
		self.fullLength[QUERY][query][ref] = self.fullLength[QUERY][query].get(ref, 0) + (queryEnd - queryStart + 1)
		self.matches[REF][ref].add(query)
		self.matches[QUERY][query].add(ref)
		self.nbParts[REF][ref]     = 0
		self.nbParts[QUERY][query] = 0
		
	def _readData(self):
		self.usedQueries = set()
		progress = UnlimitedProgress(10000, "Reading data", self.verbosity)
		for mapping in self.parser.getIterator():
			self._addMapping(mapping)
			progress.inc()
		progress.done()

	def _computeData(self):
		for type in TYPES:
			wholeCoverage   = 0
			wholeFullLength = 0
			matchesPerTe    = 0
			matchedTes      = 0
			for te in self.coverage[type]:
				message          = ""
				coverage         = len(self.coverage[type][te].values())
				wholeCoverage   += coverage
				fullLength       = 0 if not self.fullLength[type][te] else float(max(self.fullLength[type][te].values())) / self.sizes[type][te] * 100
				wholeFullLength += fullLength
				message         += "\t%s coverage of %s: %d, %d%% covered, best length: %d%%" % (TYPE_NAME[type].title(), te, coverage, float(coverage) / self.sizes[type][te] * 100, fullLength)
				matches        = len(self.matches[type][te])
				if matches > 0:
					matchesPerTe += matches
					matchedTes   += 1
					message      += " --- #matches: %d" % (matches)
				if self.verbosity > 1 and message:
				 	print message
			print "Coverage %s: %d, %d%% covered, %d%% best length --- %d matches per element" % (TYPE_NAME[type], sum(self.sizes[type].values()), float(wholeCoverage) / sum(self.sizes[type].values()) * 100, float(wholeFullLength) / len(self.coverage[type]), float(matchesPerTe) / matchedTes)
		if self.categories:
			categoryCoverage   = {}
			categorySizes      = {}
			categoryFullLength = {}
			categoryNumber     = {}
			for te in self.categories:
				categoryNumber[self.categories[te]] = categoryNumber.get(self.categories[te], 0) + 1
			for te in self.coverage[REF]:
				categoryCoverage[self.categories[te]]   = len(self.coverage[REF][te].values()) + categoryCoverage.get(self.categories[te], 0)
				categorySizes[self.categories[te]]      = self.sizes[REF][te] + categorySizes.get(self.categories[te], 0)
				categoryFullLength[self.categories[te]] = categoryFullLength.get(self.categories[te], 0) + (0 if not self.fullLength[REF][te] else float(max(self.fullLength[REF][te].values())) / self.sizes[REF][te] * 100)
			for category in categoryCoverage:
				if categorySizes[category] > 0:
					print "Coverage category %s: %d, %d%% covered" % (category, categoryCoverage[category], float(categoryCoverage[category]) / categorySizes[category] * 100)

	def _plotLandscape(self):
		for type in TYPES:
			progress = Progress(len(self.names[type]), "Plotting %s landscape" % (TYPE_NAME[type]), self.verbosity)
			for te in self.coverage[type]:
				if self.coverage[type][te]:
					plot = RPlotter("%s_landscape_%s.png" % (self.outputFileName, te.replace("/", "_")), self.verbosity)
					plot.setFill(0)
					plot.setTitle("Landscape of %s" % (te))
					plot.addLine(self.coverage[type][te])
					plot.plot()
				progress.inc()
			progress.done()

	def _plotDepth(self, coverage, type, te, wholeData):
		plot    = RPlotter("%s_depth_%s.png" % (self.outputFileName, te.replace("/", "_")), self.verbosity)
		data    = {}
		data[0] = self.sizes[type][te] - len(coverage.keys())
		for value in coverage.values():
			data[value] = data.get(value, 0) + 1
		if len(data.keys()) > 1:
			plot.setTitle("Depth of %s" % (te))
			plot.addLine(data)
			plot.plot()
		for key, value in data.iteritems():
			wholeData[key] = wholeData.get(key, 0) + value
		return wholeData

	def _plotDepthes(self):
		for type in TYPES:
			allDepthes = {}
			cpt        = 0
			progress   = Progress(len(self.names[type]), "Plotting %s depthes" % (TYPE_NAME[type]), self.verbosity)
			for te in self.coverage[type]:
				for value in self.coverage[type][te].values():
					allDepthes[cpt] = value
					cpt += 1
				allDepthes = self._plotDepth(self.coverage[type][te], type, te, allDepthes)
				progress.inc()
			progress.done()
			plot = RPlotter("%s_depth_all_%s.png" % (self.outputFileName, TYPE_NAME[type]), self.verbosity)
			plot.setTitle("Depth of all_%s" % (TYPE_NAME[type]))
			plot.addLine(allDepthes)
			plot.plot()

	def _plotCoverage(self):
		for type in TYPES:
			progress      = Progress(len(self.names[type]), "Plotting %s coverage" % (TYPE_NAME[type]), self.verbosity)
			coverages     = {}
			for te in self.coverage[type]:
				coverage       = len(self.coverage[type][te].values())
				bin            = int(float(coverage) / self.sizes[type][te] * 100 / BIN_SIZE) * BIN_SIZE
				coverages[bin] = coverages.get(bin, 0) + 1
				progress.inc()
			progress.done()
			for i in range(max(coverages.keys()) / BIN_SIZE):
				if i * BIN_SIZE not in coverages:
					coverages[i * BIN_SIZE] = 0
			plot = RPlotter("%s_%s_coverage.png" % (self.outputFileName, TYPE_NAME[type]), self.verbosity)
			plot.setTitle("% coverage")
			plot.addLine(coverages)
			plot.plot()

	def _plotMatches(self):
		for type in TYPES:
			data = {}
			for te in self.matches[type]:
				value = len(self.matches[type][te])
				data[value] = data.get(value, 0) + 1
			plot = RPlotter("%s_%s_matches.png" % (self.outputFileName, TYPE_NAME[type]), self.verbosity)
			plot.setTitle("Matches")
			plot.setFill(0)
			plot.addLine(data)
			plot.plot()

	def _plotParts(self):
		for type in TYPES:
			data = {}
			for value in self.nbParts[type].values():
				data[value] = data.get(value, 0) + 1
			plot = RPlotter("%s_%s_parts.png" % (self.outputFileName, TYPE_NAME[type]), self.verbosity)
			plot.setTitle("Parts")
			plot.setFill(0)
			plot.addLine(data)
			plot.plot()

	def _plotFullLength(self):
		for type in TYPES:
			data = dict([(i * 10, 0) for i in range(10)])
			for te in self.sizes[type]:
				fullLength          = 0 if not self.fullLength[type][te] else float(max(self.fullLength[type][te].values())) / self.sizes[type][te] * 100
				fullLengthBin       = min(99, fullLength) / 10 * 10
				data[fullLengthBin] = data.get(fullLengthBin, 0) + 1
			plot = RPlotter("%s_%s_bestLength.png" % (self.outputFileName, TYPE_NAME[type]), self.verbosity)
			plot.setTitle("Best Length")
			plot.setFill(0)
			plot.addLine(data)
			plot.plot()

	def _plot(self):
		self._plotLandscape()
		self._plotDepthes()
		self._plotCoverage()
		self._plotParts()
		self._plotMatches()
		self._plotFullLength()

	def run(self):
		self._initializeData()
		self._readData()
		self._computeData()
		if self.plots:
			self._plot()


if __name__ == "__main__":

	description = "Evaluator v1.0.1: Evaluate the results of a Blast. [Category: Personal]"

	parser = OptionParser(description = description)
	parser.add_option("-i", "--input",       dest="inputFileName",     action="store",                            type="string", help="input file [compulsory] [format: file in BLAST format]")
	parser.add_option("-o", "--output",      dest="outputFileName",    action="store",      default=None,         type="string", help="output file [format: output files]")
	parser.add_option("-r", "--reference",   dest="referenceFileName", action="store",      default=None,         type="string", help="file containing the TEs [format: file in FASTA format]")
	parser.add_option("-q", "--query",       dest="queryFileName",     action="store",      default=None,         type="string", help="file containing the assembly [format: file in FASTA format]")
	parser.add_option("-s", "--select",      dest="tesFileName",       action="store",      default=None,         type="string", help="scan only these TEs (file name with 1 TE name per line) [format: file in TXT format]")
	parser.add_option("-p", "--plot",        dest="plot",              action="store_true", default=False,                       help="also plot the data [format: boolean]")
	parser.add_option("-c", "--categories",  dest="categories",        action="store_true", default=False,                       help="also provide categories [format: boolean]")
	parser.add_option("-v", "--verbosity",   dest="verbosity",         action="store",      default=1,            type="int",    help="trace level [default: 1] [format: int]")
	(options, args) = parser.parse_args()

	e = Evaluator(options.verbosity)
	e.setPlots(options.plot)
	e.setCategories(options.categories)
	e.setInputFile(options.inputFileName)
	e.setOutputFile(options.outputFileName)
	e.setReferenceFile(options.referenceFileName)
	e.setQueryFile(options.queryFileName)
	if options.tesFileName != None:
		e.selectTes(options.tesFileName)
	e.run()

